#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

import numpy as np
from .element import Element
from .. import utils


class LineElement(Element):
    def __init__(self, *args, **kwargs):
        """
        LineElement is a class inheriting from Element.

        {
            'data': {
                'x': list[float],
                'y': list[float],
                'z': list[float],
            }
            'properties': {
                'thickness': int,
                'color': list[float],
                'alpha': float
            },
            'metadata': {
                'id': int,
                'name': str or None,
                'extension': str or None
            }
        }
        """
        super().__init__(*args, **kwargs)

    def _fill_properties(self, *args, **kwargs) -> None:
        super()._fill_properties(*args, **kwargs)
        self.thickness = kwargs.get('thickness', 1.0)
        self.loop = kwargs.get('loop', False)

    def _check_integrity(self) -> None:
        super()._check_integrity()
        if len(self.vertices) < 2:
            raise ValueError("Not enough data to create this element.")

        # Append first vertex to self.vertices if a loop was enabled
        if self.loop:
            self.vertices = np.append(self.vertices, [self.vertices[0, :]], axis=0)

    @property
    def customizable_properties(self) -> list:
        return ['color', 'alpha', 'thickness', 'loop']

    @property
    def exportable_properties(self) -> list:
        return ['color', 'alpha', 'thickness', 'loop']

    @property
    def loop(self) -> bool:
        return self.properties.get('loop')

    @property
    def thickness(self) -> int:
        return self.properties.get('thickness')

    @loop.setter
    def loop(self, _loop: bool) -> None:
        self.properties['loop'] = _loop

    @thickness.setter
    def thickness(self, value: float) -> None:
        self.properties['thickness'] = value

    """
    Utilities
    """
    def intersect_with_ray(self, origin: np.ndarray, ray: np.ndarray) -> np.ndarray:
        # Early AABB detection test
        if not utils.aabb_intersection(origin, ray, *self.bounding_box):
            return np.empty(0)

        return self.vectorized_intersection(origin, ray)

    # Adapted from https://www.codefull.net/2015/06/intersection-of-a-ray-and-a-line-segment-in-3d/
    def vectorized_intersection(self, origin: np.ndarray, ray: np.ndarray, threshold: float = 1e-2) -> np.ndarray:
        ba = np.diff(self.vertices, axis=0)
        ao = self.vertices[:-1] - origin

        ao_x_ba = np.cross(ao, ba)
        ray_x_ba = np.cross(ray, ba)

        # How much does the ray need to travel to arrive to the (infinite) line, for each segment
        s = utils.dot_by_row(ao_x_ba, ray_x_ba) / utils.magnitude2_by_row(ray_x_ba)

        # All intersections with infinite lines
        intersections = origin + s.reshape(-1, 1) * ray

        length_ai = utils.magnitude_by_row(intersections - self.vertices[:-1])
        length_ib = utils.magnitude_by_row(intersections - self.vertices[1:])
        length_ab = utils.magnitude_by_row(ba)

        # Mask is True if the ray and line are coplanar enough
        mask_threshold = abs(utils.dot_by_row(ao, ray_x_ba)) < threshold

        # Mask is True if the ray is between the start and end of each segment
        mask_inside = length_ai + length_ib <= length_ab + threshold

        return intersections[mask_threshold & mask_inside]



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

from qtpy.QtGui import QColor
from qtpy.QtWidgets import QColorDialog


class ColorDialog(QColorDialog):
    def __init__(self, element=None, parent=None):
        QColorDialog.__init__(self, parent)
        self.setOption(self.ShowAlphaChannel)
        self.setOption(self.DontUseNativeDialog)

        if bool(element):
            self.setWindowTitle(f'{self.windowTitle()} ({element.name}.{element.extension})')
            self.setCurrentColor(QColor.fromRgbF(*element.rgba))



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

import argparse
import sys


def application(paths: list) -> None:
    from qtpy.QtWidgets import QApplication
    from blastsight.view.gui.mainwindow import MainWindow

    qt_app = QApplication(sys.argv)

    w = MainWindow()
    w.show()

    # Auto-load files when called from CLI or using "Open with..." from OS
    w.viewer.load_multiple(paths, w.viewer.load_mesh)

    sys.exit(qt_app.exec_())


def container(paths: list) -> None:
    from qtpy.QtWidgets import QApplication
    from blastsight.view.gui.container import Container

    qt_app = QApplication(sys.argv)

    w = Container()

    # Activate auto-fit
    if not w.toolbar.action_collection.action_autofit.isChecked():
        w.toolbar.action_collection.action_autofit.trigger()

    w.show()

    # Auto-load files when called from CLI
    w.viewer.load_multiple(paths, w.viewer.load_mesh)

    sys.exit(qt_app.exec_())


def viewer(paths: list) -> None:
    from blastsight.view.viewer import Viewer
    v = Viewer()

    # Auto-load files when called from CLI
    v.load_multiple(paths, v.load_mesh)

    v.show()


def main() -> None:
    parser = argparse.ArgumentParser(description='Starts BlastSight in App/Container/Viewer mode.')
    parser.add_argument('-a', '--app', action='store_true', help='Starts in App mode.')
    parser.add_argument('-c', '--container', action='store_true', help='Starts in Container mode.')
    parser.add_argument('-v', '--viewer', action='store_true', help='Starts in Viewer mode.')
    parser.add_argument('paths', nargs='*', help='Paths of meshes loaded before opening BlastSight.')

    args = parser.parse_args()

    if args.viewer:
        viewer(args.paths)
    elif args.container:
        container(args.paths)
    else:
        application(args.paths)


if __name__ == '__main__':
    main()



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

import traceback

from .drawables.gldrawable import GLDrawable
from .drawables.axisgl import AxisGL
from .drawables.backgroundgl import BackgroundGL
from .drawables.meshgl import MeshGL
from .drawables.blockgl import BlockGL
from .drawables.pointgl import PointGL
from .drawables.linegl import LineGL
from .drawables.tubegl import TubeGL


class DrawableFactory:
    def __init__(self, engine):
        super().__init__()
        self.engine = engine

    @staticmethod
    def generate_drawable(d_class: type, generator: callable, *args, **kwargs) -> GLDrawable or None:
        try:
            element = generator(*args, **kwargs)
            drawable = d_class(element, *args, **kwargs)

            return drawable
        except Exception:
            traceback.print_exc()
            return None

    """
    Drawables by arguments
    """
    def axis(self, *args, **kwargs) -> AxisGL:
        return self.generate_drawable(AxisGL, self.engine.null, *args, **kwargs)

    def background(self, *args, **kwargs) -> BackgroundGL:
        return self.generate_drawable(BackgroundGL, self.engine.null, *args, **kwargs)

    def mesh(self, *args, **kwargs) -> MeshGL:
        return self.generate_drawable(MeshGL, self.engine.mesh, *args, **kwargs)

    def blocks(self, *args, **kwargs) -> BlockGL:
        return self.generate_drawable(BlockGL, self.engine.blocks, *args, **kwargs)

    def points(self, *args, **kwargs) -> PointGL:
        return self.generate_drawable(PointGL, self.engine.points, *args, **kwargs)

    def lines(self, *args, **kwargs) -> LineGL:
        return self.generate_drawable(LineGL, self.engine.lines, *args, **kwargs)

    def tubes(self, *args, **kwargs) -> TubeGL:
        return self.generate_drawable(TubeGL, self.engine.tubes, *args, **kwargs)

    """
    Drawables by path
    """
    def load_mesh(self, path: str, *args, **kwargs) -> MeshGL:
        return self.generate_drawable(MeshGL, self.engine.load_mesh, path, *args, **kwargs)

    def load_blocks(self, path: str, *args, **kwargs) -> BlockGL:
        return self.generate_drawable(BlockGL, self.engine.load_blocks, path, *args, **kwargs)

    def load_points(self, path: str, *args, **kwargs) -> PointGL:
        return self.generate_drawable(PointGL, self.engine.load_points, path, *args, **kwargs)

    def load_lines(self, path: str, *args, **kwargs) -> LineGL:
        return self.generate_drawable(LineGL, self.engine.load_lines, path, *args, **kwargs)

    def load_tubes(self, path: str, *args, **kwargs) -> TubeGL:
        return self.generate_drawable(TubeGL, self.engine.load_tubes, path, *args, **kwargs)



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

from colour import Color
from OpenGL.GL import *

from .gldrawable import GLDrawable


class BackgroundGL(GLDrawable):
    def __init__(self, element=None, *args, **kwargs):
        super().__init__(element, *args, **kwargs)
        self._top_color = [0.1, 0.2, 0.3]
        self._bottom_color = [0.4, 0.5, 0.6]

    @property
    def top_color(self) -> list:
        return self._top_color

    @property
    def bottom_color(self) -> list:
        return self._bottom_color

    @top_color.setter
    def top_color(self, color: list) -> None:
        self._top_color = color
        self.notify()

    @bottom_color.setter
    def bottom_color(self, color: list) -> None:
        self._bottom_color = color
        self.notify()

    def __dir__(self) -> list:
        return sorted(set(dir(type(self)) + list(self.__dict__.keys())))

    def draw(self) -> None:
        glDisable(GL_DEPTH_TEST)
        glDrawArrays(GL_TRIANGLES, 0, 3)
        glEnable(GL_DEPTH_TEST)



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

import pathlib
from qtpy.QtGui import QIcon


class IconCollection:
    @staticmethod
    def get(path: str) -> QIcon:
        return QIcon(f'{pathlib.Path(__file__).parent}/UI/icons/{path}')
